/*
  COPYRIGHT (c) 2012-2023 Esri R&D Center Zurich
  TRADE SECRETS: ESRI PROPRIETARY AND CONFIDENTIAL
  Unpublished material - all rights reserved under the
  Copyright Laws of the United States and applicable international
  laws, treaties, and conventions.

  For additional information, contact:
  Environmental Systems Research Institute, Inc.
  Attn: Contracts and Legal Services Department
  380 New York Street
  Redlands, California, 92373
  USA

  email: contracts@esri.com
*/

#ifndef PRT_RESOLVEMAP_H_
#define PRT_RESOLVEMAP_H_

#include "Attributable.h"
#include "Object.h"


namespace prt {


/**
 * A ResolveMap maps an arbitrary string key to an URI. Resolve maps are used for
 * asset lookups during CGA generation (e.g. @cgaref{op_i.html,i()} and
 * @cgaref{op_texture.html,texture()} operations). Use the getString() member
 * function derived from the Attributable interface to lookup a key (i.e. get it's URI).
 * Typically, a ResolveMap gets created from a RPK with the createResolveMap() function.
 * ResolveMaps can be stacked, i.e. a map can be put in place downstream.
 * @sa ResolveMapBuilder, prtx::ResolveMap, createResolveMap()
 */

class PRT_EXPORTS_API ResolveMap: public Attributable, public Object {
public:
	/**
	 * Searches the keys in the ResolveMap for the expression defined in query.
	 * Supported search queries include wildcards, regular expressions and file properties.
	 *
	 * This implements the CGA @cgaref{func_fileSearch.html,fileSearch()} function.
	 * This function is thread-safe.
	 * If the result does not fit into the allocated buffer, the result is truncated.
	 * The INOUT parameter resultSize must be compared to the original buffer size
	 * after the call to detect this situation.
	 *
	 *
	 * @param         project		The name of the current project ("projects" are "subfolders"
	 * 								in the resolve map starting with a '/').
	 * @param         query			The search query. See the CGA @cgaref{func_fileSearch.html,fileSearch()}
	 * 								function for syntax documentation.
	 *
	 * @param[in,out] result		Pointer to buffer to receive a string which contains the semicolon
	 *                              separated list of keys
	 * @param[in,out] resultSize	Pointer to size of reserved buffer; receives the size of
	 * 								the actual result (incl. terminating 0). Must be >= 1.
     * @param[out]    stat			Optional pointer to return the status.
	 * @returns result.
	 */
	virtual wchar_t* searchKey(const wchar_t* project, const wchar_t* query, wchar_t* result, size_t* resultSize, Status* stat = nullptr) const = 0;

	
	
	/** 
	 * Finds the key of the single cgb file in the ResolveMap. RPKs contain exactly one main rule file (= the cgb file) - use this function to find it.
	 * This key can then be set as an Initial Shape's ruleFile.
	 * The according URI can be obtained from the ResolveMap with getString(cgbKey). 
	 * 
	 * Note: It is possible that the CGA compiler generates more than one cgb file which are all put into the RPK and the according ResolveMap. However, only 
	 * one of them, the main cgb, can actually be used to generate models. This main cgb is found by this function. The other cgbs are essential for the main 
	 * cgb file to work but do not work on their own.  
	 * 
     * @param[out]    stat			Optional pointer to return the status.
     * @returns pointer to the key of the cgb file or nullptr in case of error. The pointer stays valid as long as the ResolveMap instance exists.
	 */ 
	virtual const wchar_t* findCGBKey(Status* stat = nullptr) const = 0;


	/**
	 * Finds all main cgb keys.
	 * While RPKs only contain exactly one rule file, ResolveMaps can in general be set to contain multiple rule files.
	 * In such cases findCGBKeys() can be used to collect all keys of main rule files.
	 *
	 *
	 * @param[in, out] result       Array to receive pointers to the keys. The pointers stay valid as long as the ResolveMap instance exists.
	 *                              If more cgbs than resultSize are found, only the first few are written to results and STATUS_BUFFER_TO_SMALL
	 *                              is written to stat.
	 * @param[in, out] resultSize   Size of the result array, will receive the number of cgbs found.
	 * @param[out]     stat			Optional pointer to return the status.
     * @returns result
	 */
	virtual const wchar_t** findCGBKeys(const wchar_t** result, size_t* resultSize, Status* stat = nullptr) const = 0;


protected:
	ResolveMap();
	virtual ~ResolveMap();
};


/**
 * Used for building a ResolveMap instance.
 */
class PRT_EXPORTS_API ResolveMapBuilder : public Object {
public:
	/**
	 * Creates a new instance which must be destroyed by the caller.
     * @param[out] status Optional pointer to return the status.
     * @returns a new ResolveMapBuilder instance or 0 on error.
	 */
	static ResolveMapBuilder* create(Status* status = nullptr);

	/**
	 * Creates a new builder instance which is initialized to the passed ResolveMap. The builder instance must be destroyed by the caller.
	 # @param      resolveMap ResolveMap to initialize this builder with.
     * @param[out] status     Optional pointer to return the status.
     * @returns a new ResolveMapBuilder instance or 0 on error.
	 */
	static ResolveMapBuilder* createFromResolveMap(ResolveMap const* resolveMap, Status* status = nullptr);

	/**
	 * Creates a new ResolveMap instance. The instance must be destroyed by the caller.
     * @param[out] status Optional pointer to return the status.
     * @returns a new ResolveMap instance or 0 on error.
	 */
	virtual ResolveMap const* createResolveMap(Status* status = nullptr) const = 0;

	/**
	 * Creates a new ResolveMap instance and resets the builder. The instance must be destroyed by the caller.
     * @param[out] status Optional pointer to return the status.
     * @returns a new ResolveMap instance or 0 on error.
	 */
	virtual ResolveMap const* createResolveMapAndReset(Status* status = nullptr) = 0;



	/**
	 * Enum for embedded entries mode in addEntry()
	 */
	enum EmbeddedEntriesMode {
		EEM_INSPECT_AND_ADD, 	///< inspect uri and add embedded resources
		EEM_IGNORE 				///< do not try to resolve and add embedded resources
	};

	/**
	 * Adds a (key, URI) entry.
	 * @param key The (unique) key string.
	 * @param uri The percent-encded URI to the (embedded) asset or texture. See \ref AppNotes for URI creation.
	 * @param embeddedEntries If mode is EEM_INSPECT_AND_ADD (default), the uri is inspected if it points to a resource for
	 *                           which a Decoder with a ResolveMapProvider which can handle embedded resources is
	 *                           registered in the ExtensionManager. If so, and the resource is inspected and embedded resources are added as
	 *                           (key, uri) pairs too.
	 *
	 * @returns status.
	 */
	virtual Status addEntry(const wchar_t* key, const wchar_t *uri, EmbeddedEntriesMode embeddedEntries = EEM_INSPECT_AND_ADD) = 0;

	/**
	 * Puts a ResolveMap in place downstream. In key lookups, the lookup propagates to the downstream map
	 * if it is not found in the current map.
	 * @param map The downstream ResolveMap to set.
	 * @returns status.
	 */
	virtual Status setDownstreamMap(const ResolveMap* map) = 0;

protected:
	ResolveMapBuilder();
	virtual ~ResolveMapBuilder();
};


} // namespace prt


#endif /* PRT_RESOLVEMAP_H_ */
